
from docido_sdk.core import Interface

__all__ = [
    'ICrawler',
]


class ICrawler(Interface):
    """Extension point interface for components willing to
    provide additional Docido crawlers.
    """
    def get_service_name(self):
        """
        :return: crawler name
        :rtype: string
        """

    def iter_crawl_tasks(index, oauth_token, logger, full):
        """Split the crawl in smaller independant actions,
        and returns them for delayed execution.

        :param docido_sdk.push.api.IndexAPI: index
          To manipulate Docido index

        :param docido_sdk.oauth.OAuthToken oauth_token:
          OAuth credentials

        :param logging.Logger logger:
          to emit messages

        :param bool full:
          whether the entire account must be pushed or only
          changes that occured since previous crawl.

        :return: a dictionary instance containing the following keys:

        - 'tasks' (mandatory): generator of :py:func:`functools.partial` tasks
          to execute to perform the account synchronization.
          partial objects may accept 3 arguments:

              - push_api (:py:class:`docido_sdk.push.IndexAPI`)
              - oauth_token (:py:class:`docido_sdk.oauth.OAuthToken`)
              - previous_result (:py:class:`object`) previous task result,
                if any.
              - logger (:py:class:`logging.Logger`)

        - 'epilogue' (optional): a :py:func:`functools.partial` instance
          to execute when all sub-tasks have been executed. The partial
          instance may accept the following arguments:

              - push_api (:py:class:`docido_sdk.push.IndexAPI`)
              - oauth_token (:py:class:`docido_sdk.oauth.OAuthToken`)
              - results (a result or a list of results)
              providing what the sub-tasks returned.
              - logger (:py:class:`logging.Logger`)

        A task cannot be instance, class, or static method
        of a :py:func:`docido_sdk.core.Component` object.
        Therefore you may provide functions defined outside
        your crawler class definition.
        """




from docido_sdk.core import (
    Component,
    ExtensionPoint,
    implements,
)
import docido_sdk.config as docido_config
from . api import (
    IndexAPIProvider,
    IndexPipelineConfig,
    PullCrawlerIndexingConfig,
)


class YamlPullCrawlersIndexingConfig(Component):
    implements(PullCrawlerIndexingConfig, IndexPipelineConfig)
    index_api_providers = ExtensionPoint(IndexAPIProvider)

    def service(self, service):
        crawlers_config = docido_config.pull_crawlers.crawlers
        if service not in crawlers_config:
            raise Exception("Cannot find config for crawler " + service)
        return crawlers_config[service].get('indexing', {})

    def core(self):
        return docido_config.pull_crawlers.indexing

    def get_pipeline(self):
        indexing_config = self.core()
        processor_pipeline = indexing_config.pipeline
        providers = dict([
            (p.__class__.__name__, p)
            for p in list(self.index_api_providers)
        ])
        return list(map(lambda p: providers[p], processor_pipeline))



__all__ = [
    'OAuthToken',
]


class OAuthToken(object):
    """OAuth credentials base-class. Several implementations are available:

    this class:
      provides crawler what is require to authenticate against sources
      via OAuth.
    """
    def __init__(self, access_token, refresh_token=None,
                 token_secret=None, consumer_key=None, expires=None):
        self.__access_token = access_token
        self.__refresh_token = refresh_token
        self.__token_secret = token_secret
        self.__consumer_key = consumer_key
        self.__expires = expires

    access_token = property(
        fget=lambda slf: slf.__access_token,
        doc='''Read-only property accessor over the
        OAuth granted access token used by Docido to gain access
        to the protected resources on behalf of the user, instead
        of using the user's service provider credentials.

        :rtype: string
        '''
    )

    refresh_token = property(
        fget=lambda slf: slf.__refresh_token,
        doc='''Read-only property accessor over the
        OAuth refresh token used to recreate the access token.

        :rtype: string
        '''
    )

    token_secret = property(
        fget=lambda slf: slf.__token_secret,
        doc='''Read-only property accessor over the
        secret token provided by a service when retrieving an OAuth token.
        This property is set only when required provided by the authentication
        mechanism of the crawled service and required by crawler to fetch data.

        :rtype: string
        '''
    )

    consumer_key = property(
        fget=lambda slf: slf.__consumer_key,
        doc='''Read-only property accessor over the
        Docido consumer key. This property is set when required by
        the crawler to fetch data.

        :rtype: string
        '''
    )

    expires = property(
        fget=lambda slf: slf.__expires,
        doc='''Read-only property accessor over the expires field provided
        by authentication mechanism of the crawled service when token was
        acquired.

        :rtype: string
        '''
    )




import functools
import inspect


class lazy(object):
    """A lazily-evaluated attribute.

    :since: 1.0
    """

    def __init__(self, fn):
        self.fn = fn
        functools.update_wrapper(self, fn)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.fn.__name__ in instance.__dict__:
            return instance.__dict__[self.fn.__name__]
        result = self.fn(instance)
        instance.__dict__[self.fn.__name__] = result
        return result

    def __set__(self, instance, value):
        instance.__dict__[self.fn.__name__] = value

    def __delete__(self, instance):
        del instance.__dict__[self.fn.__name__]


def decorate_instance_methods(obj, decorator, includes=None, excludes=None):
    """Decorator instance methods of an object.

    :param obj: Python object whose instance methods have to be decorated
    :param decorator:
      instance method decorator.
      >>> def decorate(name, f):
      >>>   def __wrap(*args, **kwargs)
      >>>     print '--> entering instance method {}'.format(name)
      >>>     eax = f(*args, **kwargs)
      >>>     print '<-- leaving instance method {}'.format(name)

    :param string list includes:
      restrict wrapped instance methods. Default is `None` meaning
      that all instance method are wrapped.
    :param string list excludes:
      used to prevent some instance methods to be wrapped. Default is `None`

    :return: new class that inherits the `clazz` specified in parameter.
    """
    class InstanceMethodDecorator(object):
        def __getattribute__(self, name):
            value = obj.__getattribute__(name)
            if excludes and name in excludes:
                return value
            if includes and name not in includes:
                return value
            if inspect.ismethod(value):
                value = decorator(name, value)
            return value
    return InstanceMethodDecorator()


def reraise(clazz):
    """ Decorator catching every exception that might be raised by wrapped
    function and raise another exception instead.
    Exception initially raised is passed in first argument of the raised
    exception.

    :param: Exception class: clazz:
      Python exception class to raise
    """
    def _decorator(f):
        @functools.wraps(f)
        def _wrap(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                raise clazz(e)
        return _wrap
    return _decorator




from contextlib import contextmanager
import copy
import os
import shutil
import tempfile


@contextmanager
def restore_dict_kv(a_dict, key, copy_func=copy.deepcopy):
    """Backup an object in a with context and restore it when leaving
    the scope.

    :param a_dict:
      associative table
    :param: key
      key whose value has to be backed up
    :param copy_func: callbable object used to create an object copy.
    default is `copy.deepcopy`
    """
    exists = False
    if key in a_dict:
        backup = copy_func(a_dict[key])
        exists = True
    try:
        yield
    finally:
        if exists:
            a_dict[key] = backup
        else:
            a_dict.pop(key, None)


@contextmanager
def unregister_component(component):
    try:
        component.unregister()
        yield component
    finally:
        component.register()


@contextmanager
def tempdir(**kwargs):
    path = tempfile.mkdtemp(**kwargs)
    try:
        yield path
    finally:
        shutil.rmtree(path)


@contextmanager
def pushd(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)




from datetime import datetime
import re

from dateutil import parser
import pytz
import six

UTC_EPOCH = datetime(1970, 1, 1).replace(tzinfo=pytz.utc)
MAX_POSIX_TIMESTAMP = pow(2, 32) - 1


class timestamp_ms(object):
    """Build UTC timestamp in milliseconds
    """

    @classmethod
    def feeling_lucky(cls, obj):
        """Tries to convert given object to an UTC timestamp is ms, based
        on its type.
        """
        if isinstance(obj, six.string_types):
            return cls.from_str(obj)
        elif isinstance(obj, six.integer_types) and obj <= MAX_POSIX_TIMESTAMP:
            return cls.from_posix_timestamp(obj)
        elif isinstance(obj, datetime):
            return cls.from_datetime(obj)
        else:
            raise ValueError("Don't know how to get timestamp")

    @classmethod
    def from_str(cls, timestr):
        """Use `dateutil` module to parse the give string
        """
        return cls.from_datetime(parser.parse(timestr))

    @classmethod
    def from_ymd(cls, year, month=1, day=1):
        return cls.from_datetime(datetime(
            year=year, month=month, day=day
        ))

    @classmethod
    def from_posix_timestamp(cls, ts):
        return cls.from_datetime(datetime.utcfromtimestamp(ts))

    @classmethod
    def from_datetime(cls, date):
        if date.tzinfo is None:
            date = date.replace(tzinfo=pytz.utc)
        seconds = (date - UTC_EPOCH).total_seconds() * 1e3
        micro_seconds = date.microsecond / 1e3
        return int(seconds + micro_seconds)

    @classmethod
    def from_imap_header(cls, date_header_value):
        if re.match(r".*\d{4}.*\d{4}", date_header_value):
            # replace '-0000' timezone information to '-00:00'
            value = re.sub(
                r"(.*)(\s+[\+-]?\d\d)(\d\d).*$",
                r"\1 \2:\3",
                date_header_value
            )
        else:
            value = date_header_value
        try:
            return cls.from_str(value)
        except ValueError:
            if re.match(r".*[\-+]?\d{2}:\d{2}$", value):
                value = re.sub(
                    r"(.*)(\s[\+-]?\d\d:\d\d)$",
                    r"\1",
                    value
                )
            return cls.from_str(value)

    @classmethod
    def now(cls):
        return cls.from_datetime(datetime.utcnow())



# encoding: utf-8
"""
Provides Python Embedded Domain Specific Languages.
"""
import operator


class kwargsql(object):
    """Query your Python objects with a `kwargs` syntax.

    Syntax looks like the Mongoengine syntax to query documents.

    Keys in the keyword argument specifies:

    - the attribute's location to test
    - the operation to perform, returning either `True` or `False`
      - first operand is the attribute value
      - second operand if the associated key value.

    For instance:
    - foo=42
    - foo__lt=43

    An extra `not` operator can be specified before the operation,
    for instance: foo__not__gt=0

    To access nested attributes, you can use the '__' separator,
    for instance "foo__bar". This syntax allows you to select:
        - an object attribute
        - value associated to a key in a `dict`.

    It is also possible to select an item of a list, by providing
    the position.


    Example:

    >>> d = {
    >>>    'foo': [
    >>>        { 'bar': 'pika' },
    >>>        { 'bar': 42 },
    >>>    ],
    >>> }
    >>> kwargsql.and_(d, foo__1__bar_gt=41, foo__size=2)
    True
    >>>

    Available operators are as follows:

    ne – not equal to
    lt – less than
    lte – less than or equal to
    gt – greater than
    gte – greater than or equal to
    not – negate a standard check, may be used before other operators
          (e.g. Q(age__not__mod=5))
    in – value is in list (a list of values should be provided)
    nin – value is not in list (a list of values should be provided)
    size – the size of the array, dict, or string is
    exists – value for field exists

    iexact – string field exactly matches value (case insensitive)
    contains – string field contains value
    icontains – string field contains value (case insensitive)
    startswith – string field starts with value
    istartswith – string field starts with value (case insensitive)
    endswith – string field ends with value
    iendswith – string field ends with value (case insensitive)

    isinstance - same as isinstance(field, value)
    issubclass - same as issubclass(field, value)
    """
    OPERATIONS = {
        'ne': operator.ne,
        'lt': operator.lt,
        'lte': operator.le,
        'gt': operator.gt,
        'gte': operator.ge,
        'in': lambda e, c: e in c,
        'nin': lambda e, c: e not in c,
        'size': lambda c, e: len(c) == e,
        'exists': lambda e, _: e is not None,
        'iexact': lambda s, e: s.lower() == e.lower(),
        'contains': lambda s, e: e in s,
        'icontains': lambda s, e: e.lower() in s.lower(),
        'startswith': lambda s, e: s.startswith(e),
        'istartswith': lambda s, e: s.lower().startswith(e.lower()),
        'endswith': lambda s, e: s.endswith(e),
        'iendswith': lambda s, e: s.lower().endswith(e.lower()),
        'isinstance': lambda v, c: isinstance(v, c),
        'issubclass': lambda v, c: issubclass(v, c),
    }

    @classmethod
    def and_(cls, obj, **kwargs):
        """Query an object

        :param obj:
          object to test

        :param kwargs: query specified in kwargssql

        :return:
          `True` if all `kwargs` expression are `True`, `False` otherwise.
        :rtype: bool
        """
        return cls.__eval_seqexp(obj, operator.and_, **kwargs)

    @classmethod
    def or_(cls, obj, **kwargs):
        """Query an object

        :param obj:
          object to test

        :param kwargs: query specified in kwargssql

        :return:
          `True` if at leat one `kwargs` expression is `True`,
          `False` otherwise.
        :rtype: bool
        """
        return cls.__eval_seqexp(obj, operator.or_, **kwargs)

    @classmethod
    def xor(cls, obj, **kwargs):
        """Query an object.

        :param obj:
          object to test

        :param kwargs: query specified in kwargssql

        :return:
          `True` if exactly one `kwargs` expression is `True`,
          `False` otherwise.
        :rtype: bool
        """
        return cls.__eval_seqexp(obj, operator.xor, **kwargs)

    @classmethod
    def get(cls, obj, expr):
        """Parse a kwargsql string expression, and return
        the target value in given object.

        Not sure if really needed, except when using kwargsql
        expressions in YAML files for instance.

        :param obj:
          navigation starting point

        :param basestring expr:
          kwargsql expression.

        :return:
          object pointed out by the expression.
        """
        return cls.__resolve_path(obj, expr.split('__'))

    @classmethod
    def _get_operation(cls, opname):
        """Get operation from its name.
        You can override this class method to provide additional
        operations.

        :param basestring opname:
          operation name

        :return:
          binary operator if found, `None` otherwise
        :rtype: callable object
        """
        return cls.OPERATIONS.get(opname)

    @classmethod
    def __resolve_path(cls, obj, path):
        path = filter(lambda s: s, path)
        if any(path):
            for attr in path:
                obj = cls._get_obj_attr(obj, attr)
        else:
            raise Exception("Nothing to do")
        return obj

    @classmethod
    def __eval_seqexp(cls, obj, op, **kwargs):
        return reduce(
            op,
            [cls._eval_exp(obj, exp, value)
             for (exp, value) in kwargs.items()]
        )

    @classmethod
    def _get_obj_attr(cls, obj, field):
        if isinstance(obj, dict):
            return obj.get(field)
        elif isinstance(obj, list):
            return obj[int(field)]
        else:
            return getattr(obj, field, None)

    @classmethod
    def _not(cls, op):
        def wrap(*args, **kwargs):
            return not(op(*args, **kwargs))
        return wrap

    @classmethod
    def _eval_exp(cls, obj, exp, value):
        op = operator.eq
        tokens = exp.split('__')[::-1]
        _op = cls._get_operation(tokens[0])
        if _op is not None:
            # this is the operator
            op = _op
            tokens = tokens[1:]
        if tokens[0] == 'not':
            op = cls._not(op)
            tokens = tokens[1:]
        return op(cls.__resolve_path(obj, reversed(tokens)), value)



import sys

from . toolbox.collections_ext import Configuration, contextobj

sys.modules[__name__] = contextobj(Configuration.from_env(
    'DOCIDO_CONFIG',
    'settings.yml',
    Configuration()
))
# This module is now the configuration object itself.




import pkg_resources
from pkg_resources import (
    working_set,
    DistributionNotFound,
    VersionConflict,
    UnknownExtra,
)

from docido_sdk.toolbox.text import exception_to_unicode


def _enable_plugin(env, module):
    """Enable the given plugin module if it wasn't disabled explicitly."""
    if env.is_component_enabled(module) is None:
        env.enable_component(module)


def load_eggs(entry_point_name):
    """Loader that loads any eggs in `sys.path`."""
    def _load_eggs(env):
        distributions, errors = working_set.find_plugins(
            pkg_resources.Environment()
        )
        for dist in distributions:
            if dist not in working_set:
                env.log.debug('Adding plugin %s from %s', dist, dist.location)
                working_set.add(dist)

        def _log_error(item, e):
            ue = exception_to_unicode(e)
            if isinstance(e, DistributionNotFound):
                env.log.debug('Skipping "%s": ("%s" not found)', item, ue)
            elif isinstance(e, VersionConflict):
                env.log.error('Skipping "%s": (version conflict "%s")',
                              item, ue)
            elif isinstance(e, UnknownExtra):
                env.log.error('Skipping "%s": (unknown extra "%s")', item, ue)
            else:
                env.log.error('Skipping "%s": %s', item,
                              exception_to_unicode(e, traceback=True))

        for dist, e in errors.iteritems():
            _log_error(dist, e)

        for entry in sorted(working_set.iter_entry_points(entry_point_name),
                            key=lambda entry: entry.name):
            env.log.debug(
                'Loading %s from %s',
                entry.name,
                entry.dist.location
            )
            try:
                entry.load(require=True)
            except Exception as e:
                _log_error(entry, e)
            else:
                _enable_plugin(env, entry.module_name)
    return _load_eggs


def load_components(env, loaders=(load_eggs('docido.plugins'),)):
    """Load all plugin components found in `sys.path`."""
    for loadfunc in loaders:
        loadfunc(env)



"""
Provides a set of Exception classes a crawler may raise when it is
not possible to consume source API because of OAuth-related issues.
"""

import sys

from .. core import DocidoError


class CrawlerError(DocidoError):
    """Common base exception class for crawl issues"""
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          reason for error.
        """
        super(CrawlerError, self).__init__(message)


class OAuthTokenExpiredError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has expired, and needs to be refreshed
    thru Docido front application.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenExpiredError, self).__init__(message)


class OAuthTokenPermanentError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has been revoked.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenPermanentError, self).__init__(message)


class OAuthTokenRefreshRequiredError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has expired or will expire
    shortly, and need to be refreshed programmaticaly.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenRefreshRequiredError, self).__init__(message)


class Retry(Exception):
    def __init__(self, kwargs=None, countdown=None, exc=None,
                 eta=None, max_retries=None):
        """Retry the task.

        This exception can be raised by a crawl task to tell the framework
        to retry it later on.

        :param kwargs:
          task keyword arguments to retry with
        :param countdown:
          time in seconds to delay the retry for.
        :param exc:
          custom exception to report when the max restart limit exceeded.
        :param eta:
          explicit time and date to run the retry for. Must be
          a :class:`~datetime.datetime` instance.
        :param max_retries:
          if set, overrides the default retry limit
        """
        self.kwargs = kwargs
        self.exc = exc
        self.countdown = countdown
        self.eta = eta
        self.max_retries = max_retries
        self.traceback = sys.exc_info[2]




from docido_sdk.core import Interface

__all__ = [
    'IndexAPI',
    'IndexAPIConfigurationProvider',
    'IndexAPIProcessor',
    'IndexAPIProvider',
    'IndexPipelineConfig',
    'PullCrawlerIndexingConfig',
]


class IndexAPIProvider(Interface):  # pragma: no cover
    """ Provide an implementation of IndexAPI
    """
    def get_index_api(self, **config):
        """ Create a new instance of :py:class:`docido_sdk.index.IndexAPI`

        :param dict: config:
          extra configuration given to the object to create

        :return: new instance
        :rtype: :py:class:`docido_sdk.index.IndexAPI`
        """


class IndexAPI(object):
    """Read/write access to Docido index.

    :An IndexAPI object can manipulate 3 kind of data:
        :cards:
          a searchable item in Docido index.
        :thumbnails:
          a binary item in Docido index, for thumbnails of
          card's attachments. Used to improve user-experience by providing
          fast preview of binary files attached to cards.
        :a key value store:
          provides crawlers a way to persist their synchronization state.

    :Error Handling:
        Every bulk operation that modifies Docido index returns the list of
        operations that failed. Every item is a `dict` providing
        the following key:

        :status:
          http error code
        :error:
          reason in string format
        :id:
          error identifier
        :card:
          original card

    :Filtering:
        Index enumeration and deletion operations allow you to restrict
        the target scope by providing a `query` in parameter.
        The `query` parameters follows the Elasticsearch Query DSL.
    """
    def push_cards(self, cards):
        """Send a synchronous bulk indexing request

        :param list cards: collections of cards to index.

        :return: collection of items whose insertion failed.
        """

    def delete_cards(self, query):
        """Send a synchronous deletion by query

        :param list query: a search definition using the Elasticsearch
            Query DSL to restrict the scope of cards to delete.

        :raise IndexAPIError: if a problem occur while deleting cards
        """

    def delete_cards_by_id(self, ids):
        """ Send a bulk deletion request

        :param list ids: a list of card ids to delete.

        :return: collection of items whose deletion failed.
        """

    def search_cards(self, query=None):
        """Enumerate cards in Docido index.

        :param list query: a search definition using the
            Elasticsearch Query DSL

        :return: FIXME
        """

    def push_thumbnails(self, thumbnails):
        """Add or update thumbnails in dedicated Docido index.

        :param list thumbnails: Collection of tuples
                                `(identifier, encoded_bytes, mime_type)`

        :return: collection of items whose insertion failed.
        """

    def delete_thumbnails(self, query):
        """Delete thumbnails from dedicated Docido index.

        :param query: a search definition using the Elasticsearch Query DSL to
                    restrict the scope of thumbnails to delete.

        :raise IndexAPIError: if a problem occur while deleting thumbnails
        """

    def delete_thumbnails_by_id(self, ids):
        """Delete thumbnails from dedicated Docido index.

        :param query: a search definition using the Elasticsearch Query DSL to
                    restrict the scope of thumbnails to delete.

        :raise IndexAPIError: if a problem occur while deleting thumbnails
        """

    def get_kv(self, key):
        """Retrieve value from persistence layer

        :param string key: input key

        :return: the value is present, `None` otherwise.
        :rtype: string
        """

    def set_kv(self, key, value):
        """Insert or update existing key in persistence layer.

        :param string key: input key
        :param string value: value to store
        """

    def delete_kv(self, key):
        """Remove key from persistent storage.

        :param key: the key to remove
        """

    def delete_kvs(self):
        """Remove all crawler persisted data.
        """

    def get_kvs(self):
        """Retrieve all crawler persisted data.

        :return: collection of tuple `(key, value)`
        :rtype: list
        """

    def ping(self):
        """Test availability of Docido index

        :raises SystemError: if Docido index is unreachable
        """


class IndexAPIProcessor(IndexAPI):
    """ Allows creation of :py:class:`docido_sdk.index.IndexAPI` pipelines
    """
    def __init__(self, parent=None, **config):
        """
        :param :py:class:`docido_sdk.index.IndexAPI`: parent:
          next pipeline object

        :param dict: config:
          extra processor configuration
        """
        self._parent = parent
        self._config = config

    def push_cards(self, cards):
        return self._parent.push_cards(cards)

    def delete_cards(self, query):
        return self._parent.delete_cards(query)

    def delete_cards_by_id(self, ids):
        return self._parent.delete_cards_by_id(ids)

    def search_cards(self, query=None):
        return self._parent.search_cards(query)

    def push_thumbnails(self, thumbnails):
        return self._parent.push_thumbnails(thumbnails)

    def delete_thumbnails(self, query):
        return self._parent.delete_thumbnails(query)

    def get_kv(self, key):
        return self._parent.get_kv(key)

    def set_kv(self, key, value):
        return self._parent.set_kv(key, value)

    def delete_kv(self, key):
        return self._parent.delete_kv(key)

    def delete_kvs(self):
        return self._parent.delete_kvs()

    def get_kvs(self):
        return self._parent.get_kvs()

    def ping(self):
        return self._parent.ping()


class IndexAPIConfigurationProvider(Interface):  # pragma: no cover
    """ An interface to provide a configuration consumed by
    :py:class:`docido_sdk.index.IndexAPIProcessor`
    """
    def get_index_api_conf(service, docido_user_id, account_login):
        """ Provides a configuration object given to every index processors

        :param: basestring: service:
          account service name (gmail, trello, ...)

        :param: basestring: docido_user_id:
          the Docido user identifier for which the IndexAPI is meant form

        :param: basestring: account_login
          the user account login for which the IndexAPI is meant for

        :return: IndexAPI Configuration
        :rtype: dict
        """
        pass


class IndexPipelineConfig(Interface):  # pragma: no cover
    """ Provides list of :py:class:`docido_sdk.index.IndexAPIProvider`
    to link together in order to create the indexing pipeline.
    """
    def get_pipeline(self):
        """
        :return:
          description of the index pipeline to create
        :rtype: :py:class:`docido_sdk.index.IndexAPIProvider`
        """
        pass


class PullCrawlerIndexingConfig(Interface):
    def core(self):
        """ Providers common pull-crawlers indexing configuration

        :return:
          description of pull-crawlers indexing Configuration
        :rtype: dict
        """

    def service(service):
        """ Provides crawler specific indexing configuration.

        :param: basestring service:

        :return:
          description of the crawler custom indexing Configuration
        :rtype: dict
        """




__all__ = [
    'activate_pyopenssl_for_urllib3',
]


def activate_pyopenssl_for_urllib3():
    """
    Workaround issue described here:
    https://urllib3.readthedocs.org/en/latest/security.html#insecureplatformwarning

    """
    import urllib3.contrib.pyopenssl
    urllib3.contrib.pyopenssl.inject_into_urllib3()



import copy
import sys

import voluptuous
from yamlious import from_dict, merge_dicts
from docido_sdk.index import (
    IndexAPIError,
    IndexAPIErrorBuilder,
    IndexAPIProcessor,
    IndexAPIProvider,
    PullCrawlerIndexingConfig,
)
from docido_sdk.core import (
    Component,
    ExtensionPoint,
    implements,
    Interface,
)
from docido_sdk.toolbox.decorators import lazy


__all__ = ['CheckProcessor']


class Check(IndexAPIProcessor):
    """ An index api processor responsible for document and query structure
    checks

    Based on predefined voluptuous Schema, every document will be checked or
    an IndexAPIError will be raised
    """
    def __init__(self, default_schema, card_schemas, query_schema, **kwargs):
        super(Check, self).__init__(**kwargs)
        assert isinstance(default_schema, voluptuous.Schema)
        for schema in card_schemas.values():
            assert isinstance(schema, voluptuous.Schema)
        assert isinstance(query_schema, voluptuous.Schema)
        self.default_schema = default_schema
        self.card_schemas = card_schemas
        self.query_schema = query_schema

    def push_cards(self, cards):
        for i in range(len(cards)):
            try:
                c = cards[i]
                if 'kind' not in c:
                    raise IndexAPIErrorBuilder(c)\
                        .message("Missing 'kind' field")\
                        .exception()
                self._check_attachments(c)
                if c['kind'] in self.card_schemas:
                    cards[i] = self.card_schemas[c['kind']](c)
                else:
                    cards[i] = self.default_schema(c)
            except voluptuous.MultipleInvalid as e:
                traceback = sys.exc_info()[2]
                raise IndexAPIErrorBuilder(c)\
                    .message(e)\
                    .exception(), None, traceback
        return super(Check, self).push_cards(cards)

    def _check_attachments(self, card):
        origin_ids = set()
        # add 'attachments' field if missing, otherwise voluptuous yells
        for attachment in card.setdefault('attachments', []):
            if 'origin_id' not in attachment:
                continue
            origin_id = attachment['origin_id']
            if origin_id in origin_ids:
                raise IndexAPIError.build(card).message(
                    "Cannot have 2 attachments with "
                    "the same 'origin_id'").exception()
            origin_ids.add(origin_id)

    def search_cards(self, query=None):
        try:
            self.query_schema(query)
        except voluptuous.MultipleInvalid as e:
            raise IndexAPIError(e)
        return super(Check, self).search_cards(query)

    def delete_thumbnails(self, query):
        try:
            self.query_schema(query)
        except voluptuous.MultipleInvalid as e:
            raise IndexAPIError(e)
        return super(Check, self).delete_thumbnails(query)

    def delete_cards(self, query):
        try:
            self.query_schema(query)
        except voluptuous.MultipleInvalid as e:
            raise IndexAPIError(e)
        return super(Check, self).delete_cards(query)


class CheckProcessorSchemaProvider(Interface):  # pragma: no cover
    def default_schema(service):
        pass

    def card_schemas(service):
        pass

    def query_schema(service):
        pass


class CheckProcessor(Component):
    implements(IndexAPIProvider)
    schema_provider = ExtensionPoint(CheckProcessorSchemaProvider, unique=True)

    def get_index_api(self, **config):
        service = config['service']
        default_schema = self.schema_provider.default_schema(service)
        query_schema = self.schema_provider.query_schema(service)
        card_schemas = self.schema_provider.card_schemas(service)
        return Check(default_schema, card_schemas, query_schema, **config)


class DocidoCheckProcessorSchemaProvider(Component):
    implements(CheckProcessorSchemaProvider)
    indexing_config = ExtensionPoint(PullCrawlerIndexingConfig, unique=True)

    def _get_config(self, indexing_config):
        check_processor = indexing_config.get('check_processor', {})
        return check_processor.get('schemas', {})

    @lazy
    def _core_config(self):
        return self._get_config(self.indexing_config.core())

    def _crawler_config(self, service):
        return self._get_config(self.indexing_config.service(service))

    def _schema_from_dicts(self, core_conf, crawler_conf):
        schema, options = from_dict(merge_dicts(
            copy.deepcopy(core_conf),
            copy.deepcopy(crawler_conf)
        ))
        return voluptuous.Schema(schema, **options)

    def _get_schemas(self, service):
        kind_schemas = self._core_config.get('card', {}).get('kind', {})
        return {
            k: self._schema_from_dicts(
                v,
                copy.deepcopy(self._crawler_config(service).get(k, {}))
            )
            for k, v in kind_schemas.iteritems()
        }

    def card_schemas(self, service):
        return self._get_schemas(service)

    def default_schema(self, service):
        core_default = self._core_config.get('card', {}).get('default', {})
        crawler_config = self._crawler_config(service).get('card', {}).get(
            'default', {}
        )
        return self._schema_from_dicts(core_default, crawler_config)

    def query_schema(self, service):
        core_query = self._core_config.get('query', {})
        crawler_query = self._crawler_config(service).get('query', {})
        return self._schema_from_dicts(core_query, crawler_query)




class TestLocalComponents:
    _components = []

    @classmethod
    def add(cls, component):
        cls._components.append(component)
        return component

    @classmethod
    def cleanup(cls):
        for component in cls._components:
            component.unregister()

cleanup_component = TestLocalComponents.add
cleanup_components = TestLocalComponents.cleanup



import copy
import errno
import os
import os.path as osp

from peak.util.proxies import ObjectWrapper

import yaml
try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader


class nameddict(dict):
    """ Provides dictionary whose keys are accessible via the property
    syntax: `obj.key`
    """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self
        self.__namify(self.__dict__)

    def __namify(self, a_dict):
        for key in a_dict.keys():
            if type(a_dict[key]) == dict:
                a_dict[key] = nameddict(a_dict[key])

    def __setitem__(self, key, value):
        if type(value) == dict:
            value = nameddict(value)
        super(nameddict, self).__setitem__(key, value)

    def __setattr__(self, key, value):
        if key != '__dict__' and type(value) == dict:
            value = nameddict(value)
        super(nameddict, self).__setattr__(key, value)

    def __deepcopy__(self, memo):
        cls = self.__class__
        content = dict()
        for k, v in self.iteritems():
            content[k] = copy.deepcopy(v, memo)
        result = cls.__new__(cls)
        result.__init__(content)
        return result


class Configuration(nameddict):
    @classmethod
    def from_file(cls, path):
        if not osp.exists(path) and not osp.isabs(path):
            path = osp.join(osp.dirname(osp.abspath(__file__)), path)
        with open(path, 'r') as istr:
            return Configuration(yaml.load(istr, Loader=Loader))

    @classmethod
    def from_env(cls, envvar, default, default_config):
        try:
            config = Configuration.from_file(os.getenv(envvar, default))
        except IOError as e:
            if e.errno == errno.ENOENT:
                config = default_config
            else:
                raise e
        return config


class contextobj(ObjectWrapper):
    """ Proxy a Python object, and provides a stack where copies of the
    wrapped object can be manipulated with `_push` and `_pop`
    member functions.

    `contextobj` can also be used in a Python with-statement.

        >>> o = contextobj(dict())
        >>> with o:
        >>>     o['foo'] = 'bar'
        >>> # the dict has been rollbacked: it is empty again
    """
    __obj_stack = None

    def __init__(self, *args, **kwargs):
        super(contextobj, self).__init__(*args, **kwargs)
        self.__obj_stack = []

    def _push(self):
        if not isinstance(self.__subject__, dict):
            raise NotImplementedError()
        self.__obj_stack.append(copy.deepcopy(self.__subject__))
        return self

    def _pop(self):
        self.__subject__.clear()
        self.__subject__.update(self.__obj_stack.pop())
        return self

    def __enter__(self):
        return self._push()

    def __exit__(self, type, value, traceback):
        self._pop()
        return False




import logging

from docido_sdk.core import *


class Environment(Component, ComponentManager):
    """Docido SDK environment manager."""

    def __init__(self):
        ComponentManager.__init__(self)
        self.log = logging.getLogger()

    def component_activated(self, component):
        """Initialize additional member variables for components.

        Every component activated through the `Environment` object
        gets an additional member variable: `env` (the environment object)
        """
        component.env = self
        super(Environment, self).component_activated(component)

    def get_index_api(self, service, user_id, account_login):
        """Provides `IndexAPI` dedicated to a crawl

        :param basestring service:
          service name
        :param basestring user_id:
          Docido user identifier
        :param basestring account_login:
          Docido service account login
        :return: index API to be used by the crawl
        :rtype: :py:class:`docido_sdk.index.IndexAPI`
        """
        # FIXME

    def setup(self):
        from docido_sdk.loader import load_components
        load_components(self)

env = Environment()





class DocidoError(Exception):
    pass


class Interface(object):
    """Marker base class for extension point interfaces."""


class ExtensionPoint(property):
    """Marker class for extension points in components."""

    def __init__(self, interface, unique=False):
        """Create the extension point.

        :param interface:
          the `Interface` subclass that defines the
          protocol for the extension point.
        :param bool unique:
          If True, then there must be exactly one `Component` implemented
          the specified `Interface` and the property value is the `Component`
          instead of the components list implementing the `Interface`.
        """
        property.__init__(self, self.extension if unique else self.extensions)
        self.interface = interface
        self.__doc__ = ("List of components that implement `~%s.%s`" %
                        (self.interface.__module__, self.interface.__name__))

    def extensions(self, component):
        """Return a list of components that declare to implement the
        extension point interface.
        """
        classes = ComponentMeta._registry.get(self.interface, ())
        components = [component.compmgr[cls] for cls in classes]
        return [c for c in components if c]

    def extension(self, component):
        components = self.extensions(component)
        if len(components) != 1:
            error = "Expected one '{interface}' component, but found {count}"
            if len(components) != 0:
                error += ': {hits}'
            hits = list(map(lambda c: c.__class__.__name__, components))
            hits.sort()
            raise Exception(error.format(
                interface=self.interface.__name__,
                count=len(components),
                hits=', '.join(hits)
            ))
        else:
            return components[0]

    def __repr__(self):
        """Return a textual representation of the extension point."""
        return '<ExtensionPoint %s>' % self.interface.__name__


class ComponentMeta(type):
    """Meta class for components.

    Takes care of component and extension point registration.
    """
    _components = set()
    _registry = {}

    def __new__(mcs, name, bases, d):
        """Create the component class."""

        new_class = type.__new__(mcs, name, bases, d)
        if name == 'Component':
            # Don't put the Component base class in the registry
            return new_class

        if d.get('abstract'):
            # Don't put abstract component classes in the registry
            return new_class

        return ComponentMeta.register(new_class)

    @classmethod
    def register(cls, component=None):
        component = component or cls
        cls._components.add(component)
        registry = cls._registry
        for cls in component.__mro__:
            for interface in cls.__dict__.get('_implements', ()):
                classes = registry.setdefault(interface, [])
                if component not in classes:
                    classes.append(component)
        return component

    @classmethod
    def unregister(cls, component):
        try:
            cls._components.remove(component)
        except:
            pass
        for interface, components in cls._registry.iteritems():
            try:
                components.remove(component)
            except:
                pass

    def __call__(cls, *args, **kwargs):
        """Return an existing instance of the component if it has
        already been activated, otherwise create a new instance.
        """
        # If this component is also the component manager, just invoke that
        if issubclass(cls, ComponentManager):
            self = cls.__new__(cls)
            self.compmgr = self
            self.__init__(*args, **kwargs)
            return self

        # The normal case where the component is not also the component manager
        msg = "First argument must be a ComponentManager instance"
        assert len(args) >= 1 and isinstance(args[0], ComponentManager), msg
        compmgr = args[0]
        self = compmgr.components.get(cls)
        # Note that this check is racy, we intentionally don't use a
        # lock in order to keep things simple and avoid the risk of
        # deadlocks, as the impact of having temporarily two (or more)
        # instances for a given `cls` is negligible.
        if self is None:
            self = cls.__new__(cls)
            self.compmgr = compmgr
            compmgr.component_activated(self)
            self.__init__()
            # Only register the instance once it is fully initialized (#9418)
            compmgr.components[cls] = self
        return self


class Component(object):
    """Base class for components.

    Every component can declare what extension points it provides, as
    well as what extension points of other components it extends.
    """
    __metaclass__ = ComponentMeta

    @staticmethod
    def implements(*interfaces):
        """Can be used in the class definition of `Component`
        subclasses to declare the extension points that are extended.
        """
        import sys

        frame = sys._getframe(1)
        locals_ = frame.f_locals

        # Some sanity checks
        msg = 'implements() can only be used in a class definition'
        assert locals_ is not frame.f_globals and '__module__' in locals_, msg

        locals_.setdefault('_implements', []).extend(interfaces)

    @classmethod
    def unregister(cls):
        cls.__class__.unregister(cls)


implements = Component.implements


class ComponentManager(object):
    """The component manager keeps a pool of active components."""

    def __init__(self):
        """Initialize the component manager."""
        self.components = {}
        self.enabled = {}
        if isinstance(self, Component):
            self.components[self.__class__] = self

    def __contains__(self, cls):
        """Return wether the given class is in the list of active
        components."""
        return cls in self.components

    def __getitem__(self, cls):
        """Activate the component instance for the given class, or
        return the existing instance if the component has already been
        activated.

        Note that `ComponentManager` components can't be activated
        that way.
        """
        if not self.is_enabled(cls):
            return None
        component = self.components.get(cls)
        if not component and not issubclass(cls, ComponentManager):
            if cls not in ComponentMeta._components:
                raise DocidoError(
                    'Component "%s" not registered' % cls.__name__
                )
            try:
                component = cls(self)
            except TypeError as e:
                raise DocidoError(
                    'Unable to instantiate component %r (%s)' % (cls, e)
                )
        return component

    def is_enabled(self, cls):
        """Return whether the given component class is enabled."""
        if cls not in self.enabled:
            self.enabled[cls] = self.is_component_enabled(cls)
        return self.enabled[cls]

    def disable_component(self, component):
        """Force a component to be disabled.

        :param component: can be a class or an instance.
        """
        if not isinstance(component, type):
            component = component.__class__
        self.enabled[component] = False
        self.components[component] = None

    def component_activated(self, component):
        """Can be overridden by sub-classes so that special
        initialization for components can be provided.
        """
        del component  # unused
        return True

    def is_component_enabled(self, cls):
        """Can be overridden by sub-classes to veto the activation of
        a component.

        If this method returns `False`, the component was disabled
        explicitly.  If it returns `None`, the component was neither
        enabled nor disabled explicitly. In both cases, the component
        with the given class will not be available.
        """
        del cls  # unused
        return True




from contextlib import contextmanager
import threading


class RWLock:
    """Synchronization object used in a solution of so-called second
    readers-writers problem. In this problem, many readers can simultaneously
    access a share, and a writer has an exclusive access to this share.
    Additionally, the following constraints should be met:
    1) no reader should be kept waiting if the share is currently opened for
        reading unless a writer is also waiting for the share,
    2) no writer should be kept waiting for the share longer than absolutely
        necessary.

    The implementation is based on [1, secs. 4.2.2, 4.2.6, 4.2.7]
    with a modification -- adding an additional lock (C{self.__readers_queue})
    -- in accordance with [2].

    Sources:
    [1] A.B. Downey: "The little book of semaphores", Version 2.1.5, 2008
    [2] P.J. Courtois, F. Heymans, D.L. Parnas:
        "Concurrent Control with 'Readers' and 'Writers'",
        Communications of the ACM, 1971 (via [3])
    [3] http://en.wikipedia.org/wiki/Readers-writers_problem
    """

    def __init__(self):
        self.__read_switch = _LightSwitch()
        self.__write_switch = _LightSwitch()
        self.__no_readers = threading.Lock()
        self.__no_writers = threading.Lock()
        self.__readers_queue = threading.Lock()
        """A lock giving an even higher priority to the writer in certain
        cases (see [2] for a discussion)"""

    def reader_acquire(self):
        self.__readers_queue.acquire()
        self.__no_readers.acquire()
        self.__read_switch.acquire(self.__no_writers)
        self.__no_readers.release()
        self.__readers_queue.release()

    def reader_release(self):
        self.__read_switch.release(self.__no_writers)

    def writer_acquire(self):
        self.__write_switch.acquire(self.__no_readers)
        self.__no_writers.acquire()

    def writer_release(self):
        self.__no_writers.release()
        self.__write_switch.release(self.__no_readers)

    @contextmanager
    def read(self):
        self.reader_acquire()
        try:
            yield
        finally:
            self.reader_release()

    @contextmanager
    def write(self):
        self.writer_acquire()
        try:
            yield
        finally:
            self.writer_release()


class _LightSwitch:
    """An auxiliary "light switch"-like object. The first thread turns on the
    "switch", the last one turns it off (see [1, sec. 4.2.2] for details)."""
    def __init__(self):
        self.__counter = 0
        self.__mutex = threading.Lock()

    def acquire(self, lock):
        self.__mutex.acquire()
        self.__counter += 1
        if self.__counter == 1:
            lock.acquire()
        self.__mutex.release()

    def release(self, lock):
        self.__mutex.acquire()
        self.__counter -= 1
        if self.__counter == 0:
            lock.release()
        self.__mutex.release()



from contextlib import contextmanager
import logging
from optparse import OptionParser
import os
import pickle
from pickle import PickleError
import sys

from .. import loader
from ..env import env
from ..oauth import OAuthToken
from ..core import (
    implements,
    Component,
    ExtensionPoint,
)
from ..crawler import ICrawler
from ..index.config import YamlPullCrawlersIndexingConfig
from ..index.processor import (
    Elasticsearch,
    CheckProcessor,
)
from docido_sdk.index.pipeline import IndexPipelineProvider
import docido_sdk.config as docido_config
from ..toolbox.collections_ext import Configuration


def oauth_tokens_from_file(full=True):
    path = os.environ.get('DOCIDO_DCC_RUNS', '.dcc-runs.yml')
    crawlers = Configuration.from_env('DOCIDO_CC_RUNS', '.dcc-runs.yml',
                                      Configuration())
    for crawler, runs in crawlers.iteritems():
        for run, run_config in runs.iteritems():
            for k in 'config', 'token':
                if k not in run_config:
                    message = ("In file {}: missing config key '{}'"
                               " in '{}/{}' crawl description.")
                    raise Exception(message.format(path, k, crawler, run))
            if 'config' not in run_config:
                raise Exception("Missing 'config' key")
            run_config.token = OAuthToken(**run_config.token)
            run_config.setdefault('full', full)
    return crawlers


class LocalRunner(Component):
    crawlers = ExtensionPoint(ICrawler)

    def _check_pickle(self, tasks):
        try:
            return pickle.dumps(tasks)
        except PickleError as e:
            raise Exception(
                'unable to serialize crawl tasks: {}'.format(str(e))
            )

    def run(self, logger, config, crawler):
        index_provider = env[IndexPipelineProvider]
        logger.info("starting crawl")
        with docido_config:
            if config.config is not None:
                docido_config.clear()
                new_config = Configuration.from_file(config.config)
                docido_config.update(new_config)
            index_api = index_provider.get_index_api(
                self.service, None, None
            )
            tasks = crawler.iter_crawl_tasks(index_api, config.token,
                                             logger, config.full)
            self._check_pickle(tasks)

            def _runtask(task, prev_result):
                return task(index_api, config.token, prev_result, logger)

            previous_result = None
            for task in tasks['tasks']:
                try:
                    previous_result = _runtask(task, previous_result)
                except Exception as e:
                    previous_result = e
            if 'epilogue' in tasks:
                _runtask(tasks['epilogue'], previous_result)

    def run_all(self, full=False):
        crawler_runs = oauth_tokens_from_file(full=full)
        for service, launches in crawler_runs.iteritems():
            self.service = service
            c = [c for c in self.crawlers if c.get_service_name() == service]
            if len(c) != 1:
                raise Exception(
                    'unknown crawler for service: {}'.format(service)
                )
            c = c[0]
            for launch, config in launches.iteritems():
                self.launch = launch
                logger = logging.getLogger(
                    '{service}.{launch}'.format(service=self.service,
                                                launch=self.launch)
                )
                self.run(logger, config, c)


def parse_options(args=None):
    if args is None:  # pragma: no cover
        args = sys.argv[1:]
    parser = OptionParser()
    parser.add_option(
        '-i',
        action='store_true',
        dest='incremental',
        help='trigger incremental crawl'
    )
    parser.add_option(
        '-v', '--verbose',
        action='count',
        dest='verbose',
        help='set verbosity level',
        default=0
    )
    return parser.parse_args(args)


def configure_loggers(verbose):  # pragma: no cover
    logging_level = logging.WARN
    if verbose == 1:
        logging_level = logging.INFO
    elif verbose > 1:
        logging.level = logging.DEBUG
    logging.basicConfig(level=logging_level)
    # shut up a bunch of loggers
    for l in [
        'elasticsearch',
        'requests.packages.urllib3.connectionpool',
        'urllib3.connectionpool',
    ]:
        logging.getLogger(l).setLevel(logging.WARNING)


def _prepare_environment(environment):
    environment = environment or env
    loader.load_components(environment)
    environment[YamlPullCrawlersIndexingConfig]
    environment[Elasticsearch]
    environment[CheckProcessor]
    environment[IndexPipelineProvider]
    from ..index.test import LocalKV, LocalDumbIndex
    environment[LocalKV]
    environment[LocalDumbIndex]
    return env


@contextmanager
def get_crawls_runner(environment=None):
    from docido_sdk.index.pipeline import IndexAPIConfigurationProvider

    class YamlAPIConfigurationProvider(Component):
        implements(IndexAPIConfigurationProvider)

        def get_index_api_conf(self, service, docido_user_id, account_login):
            return {
                'service': service,
                'docido_user_id': docido_user_id,
                'account_login': account_login
            }
    environment = _prepare_environment(environment)
    try:
        yield env[LocalRunner]
    finally:
        YamlAPIConfigurationProvider.unregister()


def run(args=None, environment=None):
    options, args = parse_options(args)
    configure_loggers(options.verbose)
    with get_crawls_runner(environment) as runner:
        runner.run_all(full=not options.incremental)




import copy
from contextlib import contextmanager
import json
import os.path as osp
import shutil
import tempfile

import six

from docido_sdk.core import (
    Component,
    implements,
)
from docido_sdk.toolbox.threading_ext import RWLock
from .api import (
    IndexAPIProcessor,
    IndexAPIProvider,
)
from .errors import IndexAPIError
from docido_sdk.toolbox.decorators import reraise


reraise = reraise(IndexAPIError)


class LocalKVProcessor(IndexAPIProcessor):
    """Local thread-safe, `IndexAPIProcessor` persistent storage
    implementation backed by a json file on the local filesystem.
    """

    __lock = RWLock()
    __store = dict()

    def __init__(self, **config):
        """
        :param path: Path to a json file where the KVS is written.
        """
        super(LocalKVProcessor, self).__init__(**config)
        local_storage = config.get('local_storage', {})
        kv_storage = local_storage.get('kv', {})
        path = kv_storage.get('path')
        if path is None:
            path = tempfile.mkdtemp(prefix='docido-local-storage-kv')
        path = osp.join(path, 'kv.yaml')
        self.__path = path
        with self.__lock.write():
            if osp.exists(path):
                with open(path) as istr:
                    self.__store = json.load(istr)

    @reraise
    def get_key(self, key):
        assert isinstance(key, six.string_types)
        with self.__lock.read():
            return self.__store.get(key)

    @reraise
    def get_kvs(self):
        with self.__lock.read():
            return copy.copy(self.__store)

    @reraise
    def set_key(self, key, value):
        assert isinstance(key, six.string_types)
        assert isinstance(value, six.string_types)
        with self.__lock.write():
            self.__store[key] = value
            self.__persist()

    @reraise
    def delete_key(self, key):
        assert isinstance(key, six.string_types)
        with self.__lock.write():
            self.__store.pop(key, None)
            self.__persist()

    @reraise
    def delete_keys(self):
        with self.__lock.write():
            self.__store.clear()
            self.__persist()

    def __persist(self):
        with open(self.__path + '.new', 'w') as ostr:
            json.dump(self.__store, ostr, indent=2)
        shutil.move(self.__path + '.new', self.__path)


class LocalKV(Component):
    implements(IndexAPIProvider)

    def get_index_api(self, **config):
        return LocalKVProcessor(**config)


class LocalDumbIndexProcessor(IndexAPIProcessor):
    """Dumb, but yet reentrant, index implementation, persisting indices
    in local-filesystem.

    Some methods does not provide all functionalities the real Docido index
    provides. More information available in documentation of the following
    member methods: `delete_cards`, `search_cards`, and `delete_thumbnails`.
    """
    __lock = RWLock()
    __cards = dict()
    __thumbnails = dict()

    def __init__(self, **config):
        super(LocalDumbIndexProcessor, self).__init__(**config)
        local_storage = config.get('local_storage', {})
        index_storage = local_storage.get('documents', {})
        path = index_storage.get('path')
        if path is None:
            path = tempfile.mkdtemp('docido-local-storage-documents')
        cards_path = osp.join(path, 'cards.yml')
        thumbnails_path = osp.join(path, 'thumbnails.yml')
        failure_probability = index_storage.get('failure_probability', 0)

        self.__cards_path = cards_path
        self.__thumbnails_path = thumbnails_path
        self.__cards = LocalDumbIndexProcessor.load_index(cards_path)
        self.__thumbnails = LocalDumbIndexProcessor.load_index(thumbnails_path)
        self.__failure_probability = failure_probability

    @contextmanager
    def __update(self, cards=False, thumbnails=False):
        self.__lock.writer_acquire()
        try:
            yield
            if cards:
                LocalDumbIndexProcessor.persist_index(
                    self.__cards, self.__cards_path
                )
            if thumbnails:
                LocalDumbIndexProcessor.persist_index(
                    self.__thumbnails,
                    self.__thumbnails_path
                )
        finally:
            self.__lock.writer_release()

    def push_cards(self, cards):
        with self.__update(cards=True):
            for card in cards:
                self.__cards[card['id']] = card

    def delete_cards(self, query=None):
        if query != {'query': {'match_all': {}}}:
            raise IndexAPIError(
                'only match_all query is currently supported, you should use' +
                ' the ElasticSearch processor along with es-settings.yml ' +
                'config file instead'
            )
        with self.__update(cards=True):
            self.__cards.clear()

    def delete_cards_by_id(self, ids):
        errors = []
        with self.__update(cards=True):
            for _id in ids:
                if _id not in self.__cards:
                    errors.append({'status': 404, 'id': _id})
                    continue
                del self.__cards[_id]
            return errors

    def search_cards(self, query=None):
        with self.__lock.read():
            fetch_fields = None
            if query and 'fields' in query.keys():
                fetch_fields = query.get('fields', None)
            result = list()
            if fetch_fields is not None:
                for card in self.__cards.values():
                    result.append(dict((k, card[k]) for k in fetch_fields))
            else:
                for card in self.__cards.values():
                    result.append(card)
        return result
        # return {
        #     'took': 1,
        #     'timed_out': False,
        #     '_shards': {
        #         'total': 1,
        #         'successful': 1,
        #         'failed': 0,
        #     },
        #     'hits': {
        #         'total': len(result),
        #         'max_score': 1.0,
        #         'hits': [{
        #             '_index': 'docido',
        #             '_type': 'item',
        #             '_id': r['id'],
        #             '_score': 1.0,
        #             '_source': r
        #         } for r in result
        #         ],
        #     }
        # }

    def push_thumbnails(self, thumbnails):
        with self.__update(thumbnails=True):
            for id_, payload, mime in thumbnails:
                self.__thumbnails[id_] = (payload, mime)

    def delete_thumbnails(self, query):
        if query != {'query': {'match_all': {}}}:
            raise IndexAPIError(
                'only match_all query is currently supported, you should use' +
                ' the ElasticSearch processor along with es-settings.yml ' +
                'config file instead'
            )
        with self.__update(thumbnails=True):
            self.__thumbnails.clear()

    def delete_thumbnails_by_id(self, ids):
        errors = []
        with self.__update(thumbnails=True):
            for _id in ids:
                if _id not in self.__thumbnails:
                    errors.append({'status': 404, 'id': _id})
                    continue
                del self.__thumbnails[_id]
            return errors

    @classmethod
    def load_index(cls, path):
        if not osp.exists(path):
            return dict()
        else:
            with open(path) as istr:
                return json.load(istr)

    @classmethod
    def persist_index(cls, index, path):
        with open(path + '.new', 'w') as ostr:
            json.dump(index, ostr, indent=2)
        shutil.move(path + '.new', path)


class LocalDumbIndex(Component):
    implements(IndexAPIProvider)

    def get_index_api(self, **config):
        return LocalDumbIndexProcessor(**config)




from . text import to_unicode


def get_last_traceback():
    """Retrieve the last traceback as an `unicode` string."""
    import traceback
    from StringIO import StringIO
    tb = StringIO()
    traceback.print_exc(file=tb)
    return to_unicode(tb.getvalue())



from elasticsearch import Elasticsearch as _Elasticsearch

from docido_sdk.core import (
    Component,
    implements,
)
from docido_sdk.index import (
    IndexAPIProcessor,
    IndexAPIProvider,
)
import docido_sdk.config as docido_config

import os

__all__ = ['Elasticsearch', 'ElasticsearchMapping']

ES_BULK_OPERATION = ['index', 'create', 'update', 'delete']


class ElasticsearchMappingProcessor(IndexAPIProcessor):
    def __init__(self, **config):
        super(ElasticsearchMappingProcessor, self).__init__(**config)
        self.update_mapping(config['service'])

    def update_mapping(self, service):
        config = docido_config.elasticsearch
        es = _Elasticsearch(
            os.getenv('ELASTICSEARCH_HOST', config.ES_HOST),
            **config.get('connection_params', {})
        )
        to_update_mappings = [
            (config.ES_INDEX, config.ES_CARD_TYPE),
        ]
        if config.get('ES_STORE_INDEX') and config.get('ES_STORE_TYPE'):
            to_update_mappings.append(
                (config.ES_STORE_INDEX, config.ES_STORE_TYPE)
            )
        for (index, doc_type) in to_update_mappings:
            self.__update_index_mapping(es, service, index, doc_type)

    @classmethod
    def __get_crawler_mapping_config(cls, service):
        pr_config = docido_config.get('pull_crawlers', {})
        crawlers_config = pr_config.get('crawlers', {})
        crawler_config = crawlers_config.get(service, {})
        crawler_index_config = crawler_config.get('indexing', {})
        crawler_es = crawler_index_config.get('elasticsearch', {})
        return crawler_es.get('mapping', {})

    def __update_index_mapping(self, es, service, index, doc_type):
        config = docido_config.elasticsearch
        index = index.format(service=service)
        doc_type = doc_type.format(service=service)
        if not es.indices.exists(index):
            es.indices.create(index)
        mappings = []
        if 'MAPPING' in config:
            if index in config.MAPPING:
                mappings.append(config.MAPPING[index])
        crawler_mapping = self.__get_crawler_mapping_config(service)
        if any(crawler_mapping):
            mappings.append(crawler_mapping)
        for mapping in mappings:
            for field in mapping.keys():
                mapping_update_response = es.indices.put_mapping(
                    index=index,
                    doc_type=doc_type,
                    body=mapping[field]
                )
                if mapping_update_response != {'acknowledged': True}:
                    raise Exception("Could not update mapping")


class ElasticsearchProcessor(IndexAPIProcessor):
    """ Main Elasticsearch entry point

    Every sent card or thumbnail will get indexed in the user's associated
    index.

    Also provide convenience method for documents search and deletion
    """

    def __init__(self, **config):
        super(ElasticsearchProcessor, self).__init__(**config)
        es_config = docido_config.elasticsearch
        service = config['service']
        fmt = {'service': service}
        self.__es_index = es_config.ES_INDEX.format(**fmt)
        self.__es_store_index = es_config.ES_STORE_INDEX.format(**fmt)
        self.__card_type = es_config.ES_CARD_TYPE.format(**fmt)
        self.__store_type = es_config.ES_STORE_TYPE.format(**fmt)
        self.__routing = config.get('elasticsearch', {}).get('routing')
        self.__es = _Elasticsearch(
            os.getenv('ELASTICSEARCH_HOST', es_config.ES_HOST),
            **es_config.get('connection_params', {})
        )
        self.__es_store = _Elasticsearch(
            os.getenv('ELASTICSEARCH_HOST', es_config.ES_HOST),
            **es_config.get('connection_params', {})
        )

    def ping(self):
        return self.__es.ping() and self.__es_store.ping()

    def search_cards(self, query):
        generated_results = 0
        batch_size = 10
        offset = 0
        body = {
            'body': query,
            'index': self.__es_index,
            'doc_type': self.__card_type,
            'size': batch_size,
            'from_': offset,
        }
        if self.__routing:
            body['routing'] = self.__routing
        search_results = self.__es.search(**body)

        while generated_results != search_results['hits']['total']:
            for hit in search_results['hits']['hits']:
                generated_results += 1
                yield hit['_source']

            offset += batch_size
            body['from_'] = offset
            search_results = self.__es.search(**body)

    def __delete_es_docs(self, body, es, index, doc_type):
        query = {
            'body': body,
            'index': index,
            'doc_type': doc_type,
        }
        if self.__routing:
            query['routing'] = self.__routing
        es.delete_by_query(**query)

    def delete_cards(self, query):
        return self.__delete_es_docs(
            query,
            self.__es,
            self.__es_index,
            self.__card_type
        )

    def delete_cards_by_id(self, ids):
        return self.__delete_by_id(ids, self.__es_index, self.__card_type)

    def delete_thumbnails(self, query):
        return self.__delete_es_docs(
            query,
            self.__es_store,
            self.__es_store_index,
            self.__store_type
        )

    def delete_thumbnails_by_id(self, ids):
        return self.__delete_by_id(ids, self.__es_store_index,
                                   self.__store_type)

    def __delete_by_id(self, ids, index, type):
        error_docs = []
        body = [{
            'delete': {
                '_index': index,
                '_type': type,
                '_id': _id
            }
        } for _id in ids]

        if not any(body):
            return error_docs

        params = {
            'body': body,
            'refresh': True,
        }
        if self.__routing:
            params['routing'] = self.__routing

        results = self.__es.bulk(**params)
        for index, result in enumerate(results['items']):
            if result['delete']['status'] is not 200:
                error_docs.append({
                    'status': result['delete']['status'],
                    'id': ids[index],
                })
        return error_docs

    def __push_es_docs(self, docs, es, index, doc_type):
        body = []
        error_docs = []

        for doc in docs:
            action = {
                'index': {
                    '_index': index,
                    '_type': doc_type
                }
            }
            body.append(action)
            body.append(doc)

        if len(body) == 0:
            return error_docs

        params = {
            'body': body,
            'refresh': True,
        }
        if self.__routing:
            params['routing'] = self.__routing

        results = es.bulk(**params)
        if results['errors']:
            for index, item in enumerate(results['items']):
                for operation in ES_BULK_OPERATION:
                    if operation in item:
                        if item[operation]['status'] not in [200, 201]:
                            error_docs.append({
                                'card': docs[index],
                                'status': item[operation]['status'],
                                'id': docs[index]['id']
                                if 'id' in docs[index] else None,
                                'error': item[operation]['error'],
                            })
                            break
        return error_docs

    def push_cards(self, cards):
        return self.__push_es_docs(
            cards,
            self.__es,
            self.__es_index,
            self.__card_type
        )

    def push_thumbnails(self, thumbnails):
        return self.__push_es_docs(
            [
                {
                    'id': t[0],
                    'content': {
                        'data': t[1],
                        'mimetype': t[2]
                    }
                }
                for t in thumbnails],
            self.__es_store,
            self.__es_store_index,
            self.__store_type
        )


class Elasticsearch(Component):
    implements(IndexAPIProvider)

    def get_index_api(self, **config):
        return ElasticsearchProcessor(**config)


class ElasticsearchMapping(Component):
    implements(IndexAPIProvider)

    def get_index_api(self, **config):
        return ElasticsearchMappingProcessor(**config)




import os


def to_unicode(text, charset=None):
    """Convert input to an `unicode` object.

    For a `str` object, we'll first try to decode the bytes using the given
    `charset` encoding (or UTF-8 if none is specified), then we fall back to
    the latin1 encoding which might be correct or not, but at least preserves
    the original byte sequence by mapping each byte to the corresponding
    unicode code point in the range U+0000 to U+00FF.

    For anything else, a simple `unicode()` conversion is attempted,
    with special care taken with `Exception` objects.
    """
    if isinstance(text, str):
        try:
            return unicode(text, charset or 'utf-8')
        except UnicodeDecodeError:
            return unicode(text, 'latin1')
    elif isinstance(text, Exception):
        if os.name == 'nt' and \
                isinstance(text, (OSError, IOError)):  # pragma: no cover
            # the exception might have a localized error string encoded with
            # ANSI codepage if OSError and IOError on Windows
            try:
                return unicode(str(text), 'mbcs')
            except UnicodeError:
                pass
        # two possibilities for storing unicode strings in exception data:
        try:
            # custom __str__ method on the exception (e.g. PermissionError)
            return unicode(text)
        except UnicodeError:
            # unicode arguments given to the exception (e.g. parse_date)
            return ' '.join([to_unicode(arg) for arg in text.args])
    return unicode(text)


def exception_to_unicode(e, traceback=False):
    """Convert an `Exception` to an `unicode` object.

    In addition to `to_unicode`, this representation of the exception
    also contains the class name and optionally the traceback.
    """
    message = '%s: %s' % (e.__class__.__name__, to_unicode(e))
    if traceback:
        from docido_sdk.toolbox import get_last_traceback
        traceback_only = get_last_traceback().split('\n')[:-2]
        message = '\n%s\n%s' % (to_unicode('\n'.join(traceback_only)), message)
    return message




import requests


def push_thumbnails_from_links(index_api, thumbnails):
    def dl_url(id_, url):
        r = requests.get(url)
        if r.status_code == 200:
            encoded_bytes = r.content.encode("base64")
            return id_, encoded_bytes, r.headers['content-type']
    index_api.add_thumbnails(map(dl_url, thumbnails))



from .. core import Component, implements, ExtensionPoint
from . api import (
    IndexAPI,
    IndexAPIConfigurationProvider,
    IndexPipelineConfig,
    IndexAPIProvider,
)


class IndexPipelineProvider(Component):
    """ This implementation of :py:class:`docido_sdk.index.IndexAPIProvider`
    provides pipelines of :py:class:`docido_sdk.index.IndexAPIProcessor`
    """
    implements(IndexAPIProvider)
    config = ExtensionPoint(
        IndexAPIConfigurationProvider,
        unique=True
    )
    pipeline = ExtensionPoint(
        IndexPipelineConfig,
        unique=True
    )

    def get_index_api(self, service, docido_user_id, account_login):
        """ Create a pipeline of :py:class:`docido_sdk.index.IndexAPIProcessor`
        based on:

        - The :py:class:`docido_sdk.index.IndexAPIConfigurationProvider`
          which provides configuration to created
          :py:class:`docido_sdk.index.IndexAPIProcessor`
        - The :py:class:`docido_sdk.index.IndexPipelineConfig` which
          provides the :py:class:`docido_sdk.index.IndexAPIProcessor`
          list to link together.

        :param basestring: service:
          account service name (gmail, trello, ...)

        :param basestring: docido_user_id:
          the Docido user identifier for which the IndexAPI is meant form

        :param basestring: account_login
          the user account login for which the IndexAPI is meant for

        :return: new pipeline of processors
        :rtype: :py:class:`docido_sdk.index.IndexAPI`
        """
        config = self.config.get_index_api_conf(
            service, docido_user_id, account_login
        )
        index_providers = self.pipeline.get_pipeline()
        index_api = IndexAPI()
        for provider in reversed(index_providers):
            index_api = provider.get_index_api(parent=index_api, **config)
        return index_api



from setuptools import setup, find_packages

module_name = 'docido-python-sdk'
root_url = 'https://github.com/cogniteev/' + module_name

__version__ = None
with open('docido_sdk/__init__.py') as istr:
    for l in filter(lambda l: l.startswith('__version__ ='), istr):
        exec(l)
__version__ = '.'.join(map(str, __version__))

setup(
    name='docido-sdk',
    version=__version__,
    description='Docido software development kit for Python',
    author='Cogniteev',
    author_email='tech@cogniteev.com',
    url=root_url,
    download_url=root_url + '/tarball/' + __version__,
    license='Apache license version 2.0',
    keywords='cogniteev docido',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],
    packages=find_packages(exclude=['*.tests']),
    test_suite='docido.sdk.test.suite',
    zip_safe=True,
    install_requires=[
        'elasticsearch==1.6.0',
        'Flask-OAuthlib>=0.8.0',
        'ProxyTypes==0.9',
        'python-dateutil>=2.4.2',
        'pytz>=2015.6',
        'setuptools>=0.6',
        'six>=1.10.0',
        'voluptuous>=0.8.7',
        'yamlious>=0.1.3',
    ],
    entry_points="""
        [console_scripts]
        dcc-run = docido_sdk.scripts.dcc_run:run
    """
)




from .. core import DocidoError


class IndexAPIError(DocidoError):
    @classmethod
    def build(cls, card):
        return IndexAPIErrorBuilder(card)


class IndexAPIErrorBuilder(object):
    def __init__(self, card, exception=None):
        """ Prepare an IndexAPIError

        :param: dict: card
          card concerned by the exception about to be raised
        """
        assert card is not None
        assert isinstance(card, dict)
        assert 'id' in card
        self.card = card
        self._message = ''
        self._exception = exception

    def message(self, message):
        """ Provides a message to the exception to be raised

        :param: basestring: message:
          additional message
        :return: this instance
        :rtype: IndexAPIErrorBuilder
        """
        self._message = message
        return self

    def exception(self):
        return IndexAPIError("in card {}: {}".format(
            self.card['id'],
            self._message)
        )
